package Mentat::Message::Storage::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Storage::Mongo - Mentat message storage in MongoDB database

=head1 SYNOPSIS

 #*******************************************************************************##
 # !!! THIS CLASS EXTENDS THE FUNCTIONALITY OF Mentat::Storage::Mongo MODULE  !!! #
 # For further usage information refer to appropriate module documentation, only  #
 # new features will be documented here.                                          #
 # Also, more usage examples may be found in Mentat.Message.Storage.Mongo.t test  #
 # script.                                                                        #
 #*******************************************************************************##

 use Mentat::Message::Storage::Mongo;

 # Instantinate the database engine
 my $engine = Mentat::Message::Storage::Mongo->new();
 my $engine = Mentat::Message::Storage::Mongo->new(connection => [host => 'host', port => 1234, timeout => 2000], database => 'db', collection => 'coll');

 # Any of the following methods may croak, if the database error is encountered
 eval {

    # Insert the Mentat::Message object to the database (argument is blessed Mentat::Message object!!!)
    my $record_id = $engine->insert_msg($mentat_message);

    # Fetch the Mentat::Message back from database (value returned is blessed Mentat::Message object!!!)
    my $mentat_message = $engine->fetch_msg($record_id);


    # Find all messages, objects returned are blessed Mentat::Message objects
    my ($mentat_messages, $count) = $engine->find_msg();

    # Find and filter, objects returned are blessed Mentat::Message objects
    my ($mentat_messages, $count) = $engine->find_msg({'_id' => $record_id1});
    my ($mentat_messages, $count) = $engine->find_msg({'a' => {'$gt' => 4}});

    # Find, filter and sort, objects returned are blessed Mentat::Message objects
    my ($mentat_messages, $count) = $engine->find_msg({'a' => {'$gt' => 4}}, {'b' => -1});

    # Find, filter, sort and limit, objects returned are blessed Mentat::Message objects
    my ($mentat_messages, $count) = $engine->find_msg({'a' => {'$gt' => 4}}, {'b' => -1},5);

    # Find, filter, sort and skip, objects returned are blessed Mentat::Message objects
    my ($mentat_messages, $count) = $engine->find_msg({'a' => {'$gt' => 4}}, {'b' => -1}, undef, 10);

    # Find, filter, sort, limit and skip, objects returned are blessed Mentat::Message objects
    my ($mentat_messages, $count) = $engine->find_msg({'a' => {'$gt' => 4}}, {'b' => -1}, 5, 10);


    # Find all messages, objects returned are blessed Mentat::Message objects. Besides
    # objects return also raw database data.
    my ($mentat_messages, $count) = $engine->find_msg_wr();

    # Find and filter, objects returned are blessed Mentat::Message objects. Besides
    # objects return also raw database data.
    my ($mentat_messages, $count) = $engine->find_msg_wr({'_id' => $record_id1});
    my ($mentat_messages, $count) = $engine->find_msg_wr({'a' => {'$gt' => 4}});

    # Find, filter and sort, objects returned are blessed Mentat::Message objects. Besides
    # objects return also raw database data.
    my ($mentat_messages, $count) = $engine->find_msg_wr({'a' => {'$gt' => 4}}, {'b' => -1});

    # Find, filter, sort and limit, objects returned are blessed Mentat::Message objects. Besides
    # objects return also raw database data.
    my ($mentat_messages, $count) = $engine->find_msg_wr({'a' => {'$gt' => 4}}, {'b' => -1},5);

    # Find, filter, sort and skip, objects returned are blessed Mentat::Message objects. Besides
    # objects return also raw database data.
    my ($mentat_messages, $count) = $engine->find_msg_wr({'a' => {'$gt' => 4}}, {'b' => -1}, undef, 10);

    # Find, filter, sort, limit and skip, objects returned are blessed Mentat::Message objects. Besides
    # objects return also raw database data.
    my ($mentat_messages, $count) = $engine->find_msg_wr({'a' => {'$gt' => 4}}, {'b' => -1}, 5, 10);


    # Example of iterative find with conversion from database document to message
    my @result = [];
    my ($cursor, $cnta) = $engine->find_i();
    # !!!PREFERED METHOD!!! Now either use internal cursor iteration with optional limit
    # This method uses the closures @result is visible in callback method
    sub cb($) {
        my $m = $engine->objectify(@_);
        push(@result, $m); # OR DO SOME OTHER STUFF HERE
    }
    my $cntb = $engine->cursor_iterate($cursor, \&cb, $limit);
    # !!!DISCOURAGED METHOD!!! Or use cursor directly
    while (my $doc = $cursor->next) {
        my $m = $engine->objectify($doc);
        push(@result, $m); # OR DO SOME OTHER STUFF HERE
    }
    # @result now contains all data

 };
 if($@)
 {
    print STDERR "MongoDB error: $@\n";
 }

 #*******************************************************************************##
 # !!! THIS CLASS EXTENDS THE FUNCTIONALITY OF Mentat::Storage::Mongo MODULE  !!! #
 # For further usage information refer to appropriate module documentation, only  #
 # new features will be documented here.                                          #
 # Also, more usage examples may be found in Mentat.Message.Storage.Mongo.t test  #
 # script.                                                                        #
 #*******************************************************************************##

=head1 DESCRIPTION

This engine is used for permanent storage of Mentat messages in document oriented
MongoDB database. It is the implementation of the base Mentat::Message::Storage::Engine
interface and can be used by Mentat::Message::Storage facade class. Implementation
is currently based on MongoDB module, which is available in CPAN repository.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), MongoDB(3), Mentat::Storage::Mongo(3), Mentat::Message::Factory(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use DateTime;
use Time::HiRes qw(time);
use Clone qw(clone);

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;
use MongoDB;
use MongoDB::BSON::Binary;

use Tie::IxHash;

#-- Custom application modules ------------------------------------------------#
use Mentat::Storage::Mongo;
use Mentat::Message::Factory;
use Value::Convertor;
use Mentat::Message::Convertor::IDEA;
use Mentat::Message::Convertor::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

use constant DFLT_CONNECTION => [ 'socket_timeout_ms' => 1000000 ];
use constant DFLT_DATABASE   => 'mentat';
use constant DFLT_COLLECTION => 'alerts';
use constant DFLT_LIMIT      => 500000;
use constant DFLT_TIMEOUT    => 1000000;

use constant KEY_ID      => '_id';
use constant KEY_TS      => 'ts';
use constant KEY_TS_UNIX => 'ts_u';
use constant KEY_RAW     => 'msg_raw';
use constant KEY_CLASS   => 'class';
use constant KEY_RAW2    => 'msg_raw2';
use constant KEY_CLASS2  => 'class2';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#
my $SINGLETON;

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Storage::Mongo');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item get() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Storage->get()
  Purpose   : Create and return reference to the singleton instance
  Arguments : Same as the the _init() method
  Throws    : Dies, if invoked on object
  Comments  : Internally uses _init() method
  See Also  : _init() method

=cut

sub get
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);

    unless ($SINGLETON) {
        $SINGLETON = $class->new(@_);
    }
    return $SINGLETON;
}

sub DESTROY
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
}

=item objectify($) [PUBLIC]

  Usage     : my $message = $storage->objectify($document);
  Purpose   : Objectify given database document into Mentat::Message object
  Arguments : HASHREF $document - [MANDATORY]
  Returns   : Mentat::Message $mentat_message
  Throws    : Dies, if not invoked on object instance

=cut

sub objectify($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($document,) = @_;

    if (exists $document->{KEY_RAW()} and exists $document->{KEY_CLASS()}) {
        return Mentat::Message::Factory->from_string($document->{KEY_RAW()}, $document->{KEY_CLASS()});
    }
    if (exists $document->{KEY_RAW2()} and exists $document->{KEY_CLASS2()}) {
        return Mentat::Message::Factory->from_string($document->{KEY_RAW2()}, $document->{KEY_CLASS2()});
    }
    else {
        return $self->_convertor_inv->convert($document, 'mongo', '(Source|Target)\[\d+\]/IP(4|6)\[\d+\]');
    }
}

=item insert_msg($) [PUBLIC]

  Usage     : my ($record_id, $error) = $storage->insert_msg($mentat_message);
  Purpose   : Insert the given Mentat::Message message
  Arguments : Mentat::Message $mentat_message - mentat message object reference
  Returns   : STRING $message_id - ID of the inserted message, on SUCCESS, undef on FAILURE
  Throws    : Dies, if not invoked on object instance
              Croaks on insert failure
  Comments  : This method dies in case of insert failure, it must be taken care of on higher levels !

=cut

sub insert_msg($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($mentat_message,) = @_;

    # Determine the current time and convert it to appropriate database format
    my $ts     = time();
    my $ts_bin = $self->_convertor->convert_ts($ts);

    # Set the storage time within the message object
    $mentat_message->path_node_set('_CESNET/StorageTime', int($ts));

    # Convert Mentat::Message to special HASH of HASHES
    my $data_tree = $self->_convertor->convert($mentat_message);
    my $hash = $data_tree->hash();

    # Add dome additional fields to the record
    $hash->{KEY_ID()}      = $mentat_message->id(); # Use the ID from Alert message
    $hash->{KEY_TS_UNIX()} = $ts;                   # Store the time of inserting into database (us unix timestamp)
    $hash->{KEY_TS()}      = $ts_bin;               # Store the time of inserting into database

    # Deprecated
    $hash->{KEY_RAW2()}   = $mentat_message->to_string();  # Store original message in string
    $hash->{KEY_CLASS2()} = $mentat_message->class;        # Store the class of the message

    print STDERR "Hash before insertion: " . Dumper($hash) if $DEVEL;

    # Try to insert Mentat message into MongoDB
    return $self->insert($hash);
}

=item fetch_msg($) [PUBLIC]

  Usage     : my $mentat_message = $storage->fetch_msg($mentat_message_id);
  Purpose   : Fetch the Mentat message with given ID from storage
  Arguments : Same as fetch() method
  Returns   : Mentat::Message $mentat_message reference on SUCCESS,
              undef if message with given ID was not found,
  Throws    : Dies, if not invoked on object instance
              Croaks on query failure
  Comments  : Internally uses fetch() method
  See Also  : fetch() method

=cut

sub fetch_msg($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    #Log::Loger->notice(__PACKAGE__, "[MMSM] Fetching single message with args: ".Dumper(@_));
    my $result = $self->fetch(@_);
    #Log::Loger->notice(__PACKAGE__, "[MMSM] Fetching single message with args DONE: ".Dumper(@_));

    if ($result) {
        #Log::Loger->notice(__PACKAGE__, "[MMSM] Objectifiing message: ".Dumper($result));
        return $self->objectify($result);
    }

    return undef;
}

=item fetch_msg_wr($) [PUBLIC]

  Usage     : my $mentat_message = $storage->fetch_msg_wr($mentat_message_id);
  Purpose   : Fetch the Mentat message with given ID from storage with raw database data
  Arguments : Same as fetch() method
  Returns   : Mentat::Message $mentat_message reference on SUCCESS,
              undef if message with given ID was not found,
  Throws    : Dies, if not invoked on object instance
              Croaks on query failure
  Comments  : Internally uses fetch() method
  See Also  : fetch() method

=cut

sub fetch_msg_wr($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    #Log::Loger->notice(__PACKAGE__, "[MMSM] Fetching single message with args: ".Dumper(@_));
    my $result = $self->fetch(@_);
    #Log::Loger->notice(__PACKAGE__, "[MMSM] Fetching single message with args DONE: ".Dumper(@_));

    if ($result) {
        #Log::Loger->notice(__PACKAGE__, "[MMSM] Objectifiing message: ".Dumper($result));
        return ($self->objectify($result), $result);
    }

    return (undef, undef);
}

=item find_msg(;$$$$) [PUBLIC]

  Usage     : my ($mentat_messages, $count) = $storage->find_msg(); or
              my ($mentat_messages, $count) = $storage->find_msg($filter);
  Purpose   : Find the Mentat message(s) according to the given filter
  Arguments : HASH REFERENCE $filter - Query filter [OPTIONAL]
              HASH REFERENCE $sort - Sorting specification [OPTIONAL]
              INTEGER $limit - Maximum number of records to be returned [OPTIONAL]
              INTEGER $skip - Number of objects to skip from the beginning of the result set [OPTIONAL]
  Returns   : ARRAY REFERENCE of Mentat::Message objects
              INTEGER total count of Mentat messages
  Throws    : Dies, if not invoked on object instance
              Same as find() method
  Comments  : Internally uses find() method
  See Also  : find() method

=cut

sub find_msg(;$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip, $quick) = @_;

    my $messages = [];

    # Find raw results
    my ($raw_messages, $count) = $self->find($filter, undef, $sort, $limit, $skip, $quick);
    map {
        # Instantinate Mentat::Message objects from the raw results
        my $mentat_message = $self->objectify($_);

        # And store it to the result
        push (@$messages, $mentat_message);
    } @$raw_messages;

    return ($messages, $count);
}

=item find_msg_wr(;$$$$) [PUBLIC]

  Usage     : my ($mentat_messages, $count) = $storage->find_msg_wr(); or
              my ($mentat_messages, $count) = $storage->find_msg_wr($filter);
  Purpose   : Find the Mentat message(s) according to the given filter with raw database data
  Arguments : HASH REFERENCE $filter - Query filter [OPTIONAL]
              HASH REFERENCE $sort - Sorting specification [OPTIONAL]
              INTEGER $limit - Maximum number of records to be returned [OPTIONAL]
              INTEGER $skip - Number of objects to skip from the beginning of the result set [OPTIONAL]
  Returns   : ARRAY REFERENCE of ARRAYS containing
                Mentat::Message objects
                HASH REFERENCE of raw database data
              INTEGER total count of Mentat messages
  Throws    : Dies, if not invoked on object instance
              Same as find() method
  Comments  : Internally uses find() method
  See Also  : find() method

=cut

sub find_msg_wr(;$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip, $quick) = @_;

    my $messages = [];

    # Find raw results
    my ($raw_messages, $count) = $self->find($filter, undef, $sort, $limit, $skip, $quick);
    map {
        # Instantinate Mentat::Message objects from the raw results
        my $mentat_message = $self->objectify($_);

        # And store it to the result along with raw data
        push (@$messages, [$mentat_message, $_]);
    } @$raw_messages;

    return ($messages, $count);
}

=item create_indexes() [PUBLIC]

  Usage     :
  Purpose   : Create indexes
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance
  Comments  :
  See Also  :

=cut

sub create_indexes()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->_create_indexes(
        [
            Tie::IxHash->new(
                KEY_TS() => 1
            ),
            {
                'name' => KEY_TS(), 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Node.Name' => 1
            ),
            {
                'name' => 'node_name', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Node.SW' => 1
            ),
            {
                'name' => 'node_sw', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Category' => 1
            ),
            {
                'name' => 'category', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Description' => 1
            ),
            {
                'name' => 'description', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Node.Name' => 1,
                'Node.SW' => 1
            ),
            {
                'name' => 'detector', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'DetectTime' => -1
            ),
            {
                'name' => 'detecttime', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Source.IP4.ip' => 1
            ),
            {
                'name' => 'source_ip4', 'background' => 1
            }
        ],
        [
            Tie::IxHash->new(
                'Source.IP4.min' => 1
            ),
            {
                'name' => 'source_ip4_min', 'background' => 1
            }
        ],
        [
            Tie::IxHash->new(
                'Source.IP4.max' => 1
            ),
            {
                'name' => 'source_ip4_max', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Target.IP4.ip' => 1
            ),
            {
                'name' => 'target_ip4', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Target.IP4.min' => 1
            ),
            {
                'name' => 'target_ip4_min', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                'Target.IP4.max' => 1
            ),
            {
                'name' => 'target_ip4_max', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                '_CESNET.ResolvedAbuses' => 1
            ),
            {
                'name' => 'resolved_abuses', 'background' => 1
            }
        ],

        [
            Tie::IxHash->new(
                '_CESNET.StorageTime' => 1
            ),
            {
                'name' => 'storage_time', 'background' => 1
            }
        ],

        #[
            #Tie::IxHash->new(
                #'Source.IP4.min' => 1,
                #'Source.IP4.max' => 1
            #),
            #{
                #'name' => 'sourcecidr', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Target.IP4.min' => 1,
                #'Target.IP4.max' => 1
            #),
            #{
                #'name' => 'targetcidr', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Source.IP4.ip' => 1,
                #'Target.IP4.ip' => 1,
                #'DetectTime' => -1,
                #'Category' => 1,
                #'Node.Name' => 1,
                #'Node.SW' => 1
            #),
            #{
                #'name' => 'source_and_target_and_detecttime_and_detector_and_class', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Source.IP4.min' => 1,
                #'Source.IP4.max' => 1,
                #'Target.IP4.min' => 1,
                #'Target.IP4.max' => 1,
                #'DetectTime' => -1,
                #'Category' => 1,
                #'Node.Name' => 1,
                #'Node.SW' => 1
            #),
            #{
                #'name' => 'sourcecidr_and_targetcidr_and_detecttime_and_detector_and_class', 'background' => 1}
        #],

        #[
            #Tie::IxHash->new(
                #'Target.IP4.ip' => 1,
                #'DetectTime' => -1,
                #'Category' => 1,
                #'Node.Name' => 1,
                #'Node.SW' => 1,
            #),
            #{
                #'name' => 'target_and_detecttime_and_detector_and_class', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Target.IP4.min' => 1,
                #'Target.IP4.max' => 1,
                #'DetectTime' => -1,
                #'Category' => 1,
                #'Node.Name' => 1,
                #'Node.SW' => 1,
            #),
            #{
                #'name' => 'targetcidr_and_detecttime_and_detector_and_class', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'DetectTime' => -1,
                #'Category' => 1,
                #'Node.Name' => 1,
                #'Node.SW' => 1,
            #),
            #{
                #'name' => 'detecttime_and_detector_and_class', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Category' => 1,
                #'Node.Name' => 1,
                #'Node.SW' => 1,
            #),
            #{
                #'name' => 'detector_and_class', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Source.IP4.ip' => 1,
                #'Target.IP4.ip' => 1,
            #),
            #{
                #'name' => 'source_and_target', 'background' => 1
            #}
        #],

        #[
            #Tie::IxHash->new(
                #'Source.IP4.min' => 1,
                #'Source.IP4.max' => 1,
                #'Target.IP4.min' => 1,
                #'Target.IP4.max' => 1

            #),
            #{'name' => 'sourcecidr_and_targetcidr', background => 1}
        #],
    );

}

#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %params = (@_) ? @_ : ();

    # Setup default values
    $params{convertor}     = Mentat::Message::Convertor::Mongo->new() unless $params{convertor};
    $params{convertor_inv} = Mentat::Message::Convertor::IDEA->new()  unless $params{convertor_inv};
    $params{connection}    = DFLT_CONNECTION unless $params{connection};
    $params{database}      = DFLT_DATABASE   unless $params{database};
    $params{collection}    = DFLT_COLLECTION unless $params{collection};
    $params{query_timeout} = DFLT_TIMEOUT    unless $params{query_timeout};

    # Store the connection parameters
    $self->{CFG_CONNECTION}    = $params{connection};
    $self->{CFG_DATABASE}      = $params{database};
    $self->{CFG_COLLECTION}    = $params{collection};
    $self->{CFG_QUERY_TIMEOUT} = $params{query_timeout};

    # Register the message convertor
    $self->_convertor($params{convertor});
    $self->_convertor_inv($params{convertor_inv});

    # Open the connection
    $self->reconnect();

    return $self;
}

=pod

=back

=cut

1;
