package Mentat::Processor::Engine::Traggregator;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Engine::Traggregator - Engine performing ALERT message storing into permanent storage

=head1 SYNOPSIS

 use Mentat::Processor::Engine::Traggregator;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Mentat::Core qw(:c_general);
use Mentat::Processor::Engine;
use Mentat::Message::Factory;
use Mentat::Storage::Factory;
use Mentat::Message::Correlator::Engine;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Processor::Engine');
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item eh_process($) [PUBLIC]

  Usage     : Emit/execute event 'message_hello' with appropriate arguments
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance or given invalid arguments

=cut

sub eh_process($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message,) = @_;
    confess "Message must be a 'Mentat::Processor::Message' implementation" unless blessed($message) and $message->isa('Mentat::Processor::Message');

    # Retrieve the ALERT message as object
    my $alert_message = $message->alert_obj();

    # Perform trhresholding and aggregation
    my @result = $self->{CORRELATOR}->correlate($alert_message);

    # Remove the message after processing
    $self->emit_event('message_remove', $message);

    my $request;
    foreach my $r (@result) {
        $request = $self->message_create();
        $request->attribute('alert_obj', $r);
        $self->emit_event('message_traggr', $request);
        Log::Loger->debug(__PACKAGE__, "[ENG:Traggregator] Correlation produced '$request' message");
    }

    return (EC_CONTINUE, $message);
}

=item eh_check_contexts() [PUBLIC]

  Usage     : Emit/execute event 'message_hello' with appropriate arguments
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance or given invalid arguments

=cut

sub eh_check_contexts(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($last_time,) = @_;
    $last_time = $last_time || time;

    my $count = 0;
    my $request;
    while (my $m = $self->{CORRELATOR}->check_contexts()) {
        $request = $self->message_create();
        $request->attribute('alert_obj', $m);
        $self->emit_event('message_traggr', $request);
        Log::Loger->debug(__PACKAGE__, "[ENG:Traggregator] Message '$request' context timeouted");
        $count++;
    }

    Log::Loger->debug(__PACKAGE__, "[ENG:Traggregator] Total of '$count' contexts timeouted");
    return (EC_CONTINUE);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($$;$) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Engine instance
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    # Call parent version first
    $self = $self->SUPER::_init(%config);

    die "[ENG:Traggregator] Constructor expects configurations as argument\n" unless %config;
    die "[ENG:Traggregator] Engine configuration 'filter_grammar_file' is missing\n" unless $config{filter_grammar_file};
    die "[ENG:Traggregator] Engine configuration 'traggregator_config' is missing\n" unless $config{traggregator_config};

    $self->{CORRELATOR} = Mentat::Message::Correlator::Engine->new($config{filter_grammar_file}, $config{traggregator_config});

    Log::Loger->info(__PACKAGE__, "[ENG:Traggregator] Using file '$config{filter_grammar_file}' as filter grammar source");

    return $self;
}

# _get_events_handled(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of events and appropriate handlers for this component
#  Arguments : LIST OF ARRAY REFERENCES @lists - event handler lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $handlers with internal structure [$event, $handler]
#                  STRING $event   - event name
#                  STRING $handler - event handler name
#  Throws    : Dies, if not invoked on instance

sub _get_events_handled(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;

    return $self->SUPER::_get_events_handled([
        ['process', 'eh_check_contexts'],
    ], @lists);
}

=pod

=back

=cut

1;
